//          __        ____ ____
// ___  ___/  |_ __ _/_   /_   |
// \  \/ /\   __\  |  \   ||   |
//  \   /  |  | |  |  /   ||   |
//   \_/   |__| |____/|___||___|
//
//  License: BSD License ; see LICENSE
//

#ifndef VTU11_TESTING
#define VTU11_TESTING

#include "catch2/catch.hpp"

#include <string>

namespace vtu11testing
{

std::string readFile( const std::string& filename );

} // namespace vtu11testing

#endif // VTU11_TESTING
